import '../core/build_context.dart';
import '../core/store.dart';
import 'inherited_widget.dart';

/// [Provider] utilities that are added to [BuildContext].
extension ProviderHelpers on BuildContext {
  /// Returns the first [Store] with the exact type [T] provided by the
  /// nearest parent [Provider].
  ///
  /// Also, if the [Provider] is updated, the widget owning this
  /// context is rebuilt.
  T dependOnProvidedStoreOfExactType<T extends Store>() => Provider.of(this)
      .stores
      .firstWhere((final store) => store.runtimeType == T) as T;
}

/// A widget that propagates multiple [Store] instances down the tree.
///
/// **Notice:** It's the developer's responsibility to handle the initialization
/// and disposal of stores by calling [Store.initialize] and [Store.dispose]
/// whenever needed.
final class Provider extends InheritedWidget {
  final List<Store> stores;

  /// Creates a new [Provider].
  const Provider({
    required this.stores,
    required super.child,
    super.key,
    super.ref,
  });

  factory Provider.of(final BuildContext context) =>
      context.dependOnInheritedWidgetOfExactType<Provider>();

  @override
  updateShouldNotify(final Provider oldWidget) => stores != oldWidget.stores;
}
